import { BaseEntity, Connection, EntityTarget, FindManyOptions, FindOneOptions, getConnection, ObjectID, QueryRunner, SaveOptions } from "typeorm";

export class BaseRepository {
    context: IContext

    constructor(ctx: IContext | null = null) {
        if (ctx) {
            this.context = ctx
        }
    }

    async resolveContext() {

        if (!this.context || !this.context.connection) {

            const conn = getConnection()
            const runner = conn.createQueryRunner()

            this.context = {
                connection: conn,
                runner: runner
            }

            await this.context.runner.connect();

        }

        if (!this.context.runner.isTransactionActive)
            await this.context.runner.startTransaction();

    }

    async getContext() {
        return this.context
    }

    async findOneOrFail<Entity>(entityClass: EntityTarget<Entity>, id?: string | number | Date | ObjectID, options?: FindOneOptions<Entity>): Promise<Entity> {
        this.resolveContext()

        if (!this.context.runner) {
            throw new Error(NO_CONTEXT_ERROR)
        }

        return await this.context.runner.manager.findOneOrFail(entityClass, id)
    }

    async findOne<Entity>(entityClass: EntityTarget<Entity>, id?: string | number | Date | ObjectID, options?: FindOneOptions<Entity>): Promise<Entity | undefined> {
        this.resolveContext()

        if (!this.context.runner) {
            throw new Error(NO_CONTEXT_ERROR)
        }

        return await this.context.runner.manager.findOne(entityClass, id, options)
    }

    async find<Entity>(entityClass: EntityTarget<Entity>, options?: FindManyOptions<Entity>): Promise<Entity[]> {
        this.resolveContext()

        if (!this.context.runner) {
            throw new Error(NO_CONTEXT_ERROR)
        }

        return await this.context.runner.manager.find(entityClass, options)
    }

    async save<Entity>(entity: Entity, options?: SaveOptions): Promise<Entity> {
        this.resolveContext()

        if (!this.context.runner) {
            throw new Error(NO_CONTEXT_ERROR)
        }

        return await this.context.runner.manager.save(entity, options)
    }

    async saveArray<Entity>(entities: Entity[], options?: SaveOptions): Promise<Entity[]> {
        this.resolveContext()
        return await this.context.runner.manager.save(entities, options)
    }

    async commitTransaction(): Promise<void> {

        this.resolveContext()

        if (!this.context.runner) {
            throw new Error(NO_CONTEXT_ERROR)
        }

        await this.context.runner.commitTransaction()
        await this.context.runner.release()

        this.context = null
    }

    async rollbackTransaction(): Promise<void> {
        this.resolveContext()

        if (!this.context.runner) {
            throw new Error(NO_CONTEXT_ERROR)
        }

        await this.context.runner.rollbackTransaction()
        await this.context.runner.release()

        this.context = null
    }

}

export interface IContext {
    connection: Connection
    runner: QueryRunner
}

const NO_CONTEXT_ERROR = 'No Context!'
